/*

  SIMEV 0.1 - a SIMulation of EVolution

  Danilo Poccia 1998.04.15
  
*/

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <unistd.h>
#include <time.h>
#include <math.h>
#include <string.h>
#include <grdriver.h>
#include <grx20.h>
#include <curses.h>

#define NUM_COLORS                            256
#define MIN_COLOR                             127
#define STEP_COLOR                             13
#define X_SCR                                1024
#define Y_SCR                                 768
#define X_DIM                                 400
#define Y_DIM                                 400
#define X_FIX                                  10
#define Y_FIX                                  18
#define X_TEXT_FIX                (X_DIM+2*X_FIX)
#define X_TEXT_2_FIX          (X_DIM+2*X_FIX+520)
#define Y_TEXT_FIX                          Y_FIX
#define Y_TEXT_2_FIX              (Y_DIM+Y_FIX+3)
#define X_TEXT                                 80
#define Y_TEXT                                 25

#define MAX_ENERGY                            255
#define INC_ENERGY                              1
#define START_ENERGY                           10
#define REP_ENERGY                             20
#define REP_GEN_ENERGY                          0
#define MOV_ENERGY                              0
#define ATK_ENERGY                              1
#define ATK_GET_ENERGY                          5
#define LXX_ENERGY                              1
#define CXX_ENERGY                              1
#define VIR_ENERGY                              1
#define IMM_ENERGY                              1
#define MUT_PROB                              100	/* in 1/1000 */
#define ENE_PROB                             1000	/* in 1/1000 */
#define MAX_AGE                               100
#define MAX_GEN_AGE                             0
#define MAX_DIST                                3

#define BASM_DIM                               55
#define MAX_DIM                                64
#define MIN_NUM_SAVE_GEN                     1000
#define NUM_BEST_GENOMA                        30
#define NUM_OPTIONS                             7

const int dx[9] =
{0, 1, 1, 1, 0, -1, -1, -1, 0};
const int dy[9] =
{0, -1, 0, 1, 1, 1, 0, -1, -1};

const char ch[] =
"ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz";

const char basm[BASM_DIM][4] =
{"NOP",
 "MOV",
 "ENE",
 "REP",
 "INF",
 "LFN", "LFF", "LFD", "LFS", "LFB", "LFE", "LEI", "LDI", "LFY", "LFO",
 "CFN", "CFF", "CFD", "CFS", "CFB", "CFE", "CEI", "CDI", "CFY", "CFO",
 "RND",
 "IF0", "IFN", "IFE", "IFD",
 "TRN", "TRR", "TRL",
 "GOF", "GOB",
 "TO0", "TO1", "TO2", "TO3", "TO4", "TO5", "TO6", "TO7", "TO8",
 "EOT",
 "SWP", "WME", "RME",
 "WSI", "RSI",
 "VIR", "IMM",
 "ATK",
 "ELS",
 "EIF",};


typedef unsigned char a_gen;

typedef struct struct_a_genoma
  {
    unsigned long int id;
    unsigned long int prev_id;
    unsigned long int birth_time;
    unsigned long int num;
    unsigned long int max_num;
    signed char dim;
    a_gen gen[MAX_DIM];
    void *prev_genoma;
    void *next_genoma;
    unsigned long int num_sup_genoma;
  }
a_genoma;

typedef struct struct_a_entity
  {
    signed int energy;
    signed char mem;
    signed char reg;
    a_genoma *genoma;
    a_genoma *sub_genoma;
    signed int sub_prob;
    signed char gen_time;
    unsigned int time;
    unsigned char status;
  }
a_entity;

typedef struct struct_a_world
  {
    a_entity *entity;
    signed char signal;
  }
a_world;

typedef struct struct_a_coord
  {
    int x;
    int y;
    void *prev_coord;
    void *next_coord;
  }
a_coord;


a_genoma *first_genoma, *last_genoma, *genoma, *new_genoma, *best_genoma[NUM_BEST_GENOMA + 1];

a_world world[Y_DIM][X_DIM];

a_entity *entity, *new_entity, *mut_entity;

a_coord *head_coord, *first_coord, *last_coord, *coord, *new_coord;

int x, y, nx, ny, d0, d, dist, dir, pos, j, i, show_world = 1, show_stat = 1, single_step_mode = 0;

char wk_mem;

void *wk_ptr;

long unsigned int num_pop, num_born, num_dead, old_pop, num_active_genomi, num_genomi;

long int num_dim[MAX_DIM];

int max_dim, min_dim;

long int max_genoma_id = 0, num_best_genoma[NUM_BEST_GENOMA + 1];

char c;

unsigned long int world_num = 0, world_time = 0;

int eot, evolve_count;

FILE *genfile;

char text[200];

int line;

int update_stat_rate = 1000, count_stat_rate;

unsigned char num_ene, num_mov, num_atk, num_vir, num_imm;

char type_id[6];

int color_type = 0, color;

int inc_energy = INC_ENERGY,
  start_energy = START_ENERGY,
  rep_energy = REP_ENERGY,
  rep_gen_energy = REP_GEN_ENERGY,
  mov_energy = MOV_ENERGY,
  atk_energy = ATK_ENERGY,
  atk_get_energy = ATK_GET_ENERGY,
  lxx_energy = LXX_ENERGY,
  cxx_energy = CXX_ENERGY,
  vir_energy = VIR_ENERGY,
  imm_energy = IMM_ENERGY,
  mut_prob = MUT_PROB,
  ene_prob = ENE_PROB,
  max_age = MAX_AGE,
  max_gen_age = MAX_GEN_AGE,
  max_dist = MAX_DIST;

const char option_text[][20] = {
  "INC_ENERGY",
  "START_ENERGY",
  "REP_ENERGY",
  "REP_GEN_ENERGY",
  "MOV_ENERGY",
  "ATK_ENERGY",
  "ATK_GET_ENERGY" };

int *option_value [] = {
  &inc_energy,
  &start_energy,
  &rep_energy,
  &rep_gen_energy,
  &mov_energy,
  &atk_energy,
  &atk_get_energy };


void
set_up_text_mode ()
{
  GrSetMode (GR_width_height_text, X_TEXT, Y_TEXT);
}

void
clear_graph_mode ()
{
  GrClearScreen (GrBlack ());
  GrBoxNC (0, 0, X_SCR - 1, Y_SCR - 1, GrWhite ());
  GrBoxNC (X_FIX - 1, Y_FIX - 1, X_FIX + X_DIM, Y_FIX + Y_DIM, GrWhite ());
  strcpy (text, "SIMEV 0.1 - [W]ORLD  [S]TAT  [1]..[7] STAT UPDATE RATE  [R]EFRESH  [C]OLOR  [P] STEP MODE  [ ]STEP  [H]ELP  [O]PTIONS  [Q]UIT");
  GrTextXY (X_FIX, 2, text, GrWhite (), GrBlack ());
}

void
set_up_graph_mode ()
{
  GrSetMode (GR_width_height_color_graphics, X_SCR, Y_SCR, NUM_COLORS);
  GrSetRGBcolorMode ();
  clear_graph_mode ();
}

int
rnd (int max_num)
{
  return (rand () % max_num);
}

void
del_genoma (a_genoma * d_genoma)
{

  if (d_genoma->max_num >= MIN_NUM_SAVE_GEN)
    {
      fprintf (genfile, "%10lu %8lu %8lu %8lu %8lu %8lu %2d",
	       world_time,
	       d_genoma->id, d_genoma->prev_id,
	       d_genoma->birth_time,
	       d_genoma->max_num, d_genoma->num,
	       d_genoma->dim);
      for (j = 0; j < d_genoma->dim; j++)
	fprintf (genfile, " %3s", basm[d_genoma->gen[j]]);
      fprintf (genfile, "\n");
    }
  
  if (d_genoma == first_genoma)
    first_genoma = d_genoma->next_genoma;
  else
    ((a_genoma *) d_genoma->prev_genoma)->next_genoma =
      d_genoma->next_genoma;
  
  if (d_genoma == last_genoma)
    last_genoma = d_genoma->prev_genoma;
  else
    ((a_genoma *) d_genoma->next_genoma)->prev_genoma =
      d_genoma->prev_genoma;
  
  free (d_genoma);
}

void
free_mem ()
{
  genoma = first_genoma;
  do
    {
      wk_ptr = genoma->next_genoma;
      del_genoma (genoma);
      genoma = wk_ptr;
    }
  while (genoma);

  free (head_coord);

  coord = first_coord;
  do
    {
      free (world[coord->y][coord->x].entity);
      wk_ptr = coord->next_coord;
      free (coord);
      coord = wk_ptr;
    }
  while (coord);
}


void
exit_simev (int code)
{
  free_mem ();
  fclose (genfile);
  exit (code);
}

void
add_coord (int y, int x)
{
  if (x > X_DIM)
    exit_simev(12);
  if (y > Y_DIM)
    exit_simev(13);

  if ((new_coord = malloc (sizeof (a_coord))) == NULL)
    {
      set_up_text_mode ();
      printf ("End of memory...\n\n");
      exit_simev (1);
    }

  new_coord->y = y;
  new_coord->x = x;

  new_coord->next_coord = coord;
  new_coord->prev_coord = coord->prev_coord;

  ((a_coord *) coord->prev_coord)->next_coord = new_coord;
  coord->prev_coord = new_coord;

  if (first_coord == coord)
    first_coord = new_coord;
}

void
del_curr_coord ()
{
  ((a_coord *) coord->prev_coord)->next_coord = coord->next_coord;

  if (coord == first_coord)
    first_coord = coord->next_coord;

  if (coord == last_coord)
    last_coord = coord->prev_coord;
  else
    ((a_coord *) coord->next_coord)->prev_coord = coord->prev_coord;

  wk_ptr = coord->prev_coord;
  free (coord);
  coord = wk_ptr;
}

void *
get_near (int nd, int dist)
{
  if (nd < 0)
    exit_simev(10);
  if (nd > 8)
    exit_simev(11);
  nx = x + dx[nd] * dist;
  ny = y + dy[nd] * dist;
  if (nx < 0)
    nx += X_DIM;
  else if (nx >= X_DIM)
    nx -= X_DIM;
  if (ny < 0)
    ny += Y_DIM;
  else if (ny >= Y_DIM)
    ny -= Y_DIM;
  if (world[ny][nx].entity == NULL)
    {
      if ((world[ny][nx].entity = malloc (sizeof (a_entity))) == NULL)
	{
	  set_up_text_mode ();
	  printf ("End of memory...\n\n");
	  exit_simev (1);
	}
      world[ny][nx].entity->energy = 0;
      world[ny][nx].entity->status = 0;
      world[ny][nx].signal = 0;
      add_coord (ny, nx);
    }
  return world[ny][nx].entity;
}

void *
get_near_full (int nd, int dist)
{
  if (nd < 0)
    exit_simev(10);
  if (nd > 8)
    exit_simev(11);
  nx = x + dx[nd] * dist;
  ny = y + dy[nd] * dist;
  if (nx < 0)
    nx += X_DIM;
  else if (nx >= X_DIM)
    nx -= X_DIM;
  if (ny < 0)
    ny += Y_DIM;
  else if (ny >= Y_DIM)
    ny -= Y_DIM;
  if ((world[ny][nx].entity)
      && (world[ny][nx].entity->energy == 0))
    return NULL;
  else
    return world[ny][nx].entity;
}

void *
get_near_empty (int nd, int dist)
{
  if (nd < 0)
    exit_simev(10);
  if (nd > 8)
    exit_simev(11);
  nx = x + dx[nd] * dist;
  ny = y + dy[nd] * dist;
  if (nx < 0)
    nx += X_DIM;
  else if (nx >= X_DIM)
    nx -= X_DIM;
  if (ny < 0)
    ny += Y_DIM;
  else if (ny >= Y_DIM)
    ny -= Y_DIM;
  if (world[ny][nx].entity == NULL)
    {
      if ((world[ny][nx].entity = malloc (sizeof (a_entity))) == NULL)
	{
	  set_up_text_mode ();
	  printf ("End of memory...\n\n");
	  exit_simev (1);
	}
      world[ny][nx].entity->energy = 0;
      world[ny][nx].entity->status = 0;
      world[ny][nx].signal = 0;
      add_coord (ny, nx);
      return world[ny][nx].entity;
    }
  else if (world[ny][nx].entity->energy > 0)
    return NULL;
  else
    return world[ny][nx].entity;
}


void
show_option (int option)
{
  sprintf (text, "%2d - %20s : %4d          ",
	   option + 1, option_text[option], *option_value[option]);
  GrTextXY (X_FIX, Y_FIX + 20*option, text, GrWhite (), GrBlack ());
}  


void
ask_option (int option)
{
  int answer_pos = 0;
  char question_text[80], answer_text[80];
  
  sprintf (question_text, "%2d - %20s [%4d] : ",
	   option + 1, option_text[option], *option_value[option]);
  answer_text[0] = 0;
  do
    {
      strcpy (text, question_text);
      strcat (text, answer_text);
      strcat (text, "_ ");
      GrTextXY (X_FIX, Y_FIX + 20*option, text, GrWhite (), GrBlack ());
      do
	c = getch();
      while (c == ERR);
      if ((isdigit(c)) && (answer_pos < 40))
	{
	  answer_text[answer_pos++] = c;
	  answer_text[answer_pos] = 0;
	}
      else if (c == 8)
	answer_text[--answer_pos] = 0;
    }
  while (c != 10);

  strcpy (text, question_text);
  strcat (text, answer_text);
  strcat (text, " ");
  GrTextXY (X_FIX, Y_FIX + 20*option, text, GrWhite (), GrBlack ());
  
  if (answer_text[0] != 0)
    *option_value[option] = atoi(answer_text);
}  


void
init_world ()
{
  memset (world, 0, sizeof (world));
}

void
fill_world ()
{
  x = X_DIM / 2;
  y = Y_DIM / 2;

  if ((world[y][x].entity = malloc (sizeof (a_entity))) == NULL)
    {
      set_up_text_mode ();
      printf ("End of memory...\n\n");
      exit_simev (1);
    }

  entity = world[y][x].entity;

  entity->energy = START_ENERGY;
  entity->mem = 0;
  entity->reg = 0;
  entity->gen_time = 0;
  entity->time = 0;
  entity->status = 2;
  entity->sub_genoma = NULL;
  entity->sub_prob = 0;

  if ((genoma = malloc (sizeof (a_genoma))) == NULL)
    {
      set_up_text_mode ();
      printf ("End of memory...\n\n");
      exit_simev (1);
    }

  entity->genoma = genoma;

  first_genoma = genoma;
  first_genoma->prev_genoma = NULL;

  last_genoma = genoma;
  last_genoma->next_genoma = NULL;

  genoma->id = ++max_genoma_id;
  genoma->prev_id = 0;
  genoma->birth_time = 0;
  genoma->num = 1;
  genoma->max_num = 1;
  genoma->dim = 2;
  genoma->gen[0] = 2;
  genoma->gen[1] = 3;

  genoma->num_sup_genoma = 0;

  if ((head_coord = malloc (sizeof (a_coord))) == NULL)
    {
      set_up_text_mode ();
      printf ("End of memory...\n\n");
      exit (1);
    }

  head_coord->y = 0;
  head_coord->x = 0;
  head_coord->prev_coord = NULL;

  if ((first_coord = malloc (sizeof (a_coord))) == NULL)
    {
      set_up_text_mode ();
      printf ("End of memory...\n\n");
      exit (1);
    }

  first_coord->y = y;
  first_coord->x = x;
  first_coord->prev_coord = head_coord;
  first_coord->next_coord = NULL;

  head_coord->next_coord = first_coord;

  last_coord = first_coord;

}

void
get_color (a_genoma * genoma)
{
  switch (color_type)
    {
    case 0:
      color = MIN_COLOR +
	((genoma->id * STEP_COLOR) % (NUM_COLORS - MIN_COLOR));
      break;
    case 1:
      color = MIN_COLOR +
	((genoma->dim * STEP_COLOR) % (NUM_COLORS - MIN_COLOR));
      break;
    }
}

void
show_grx_world_all ()
{
  for (y = 0; y < Y_DIM; y++)
    for (x = 0; x < X_DIM; x++)
      if (world[y][x].entity)
	{
	  entity = world[y][x].entity;
	  if (entity->energy > 0)
	    {
	      get_color (entity->genoma);
	      GrPlotNC ((X_FIX + x), (Y_FIX + y), color);
	    }
	  else
	    GrPlotNC ((X_FIX + x), (Y_FIX + y), GrBlack ());
	}
      else
	GrPlotNC ((X_FIX + x), (Y_FIX + y), GrBlack ());
}

void
show_grx_world ()
{
  for (coord = first_coord; coord; coord = coord->next_coord)
    {
      y = coord->y;
      x = coord->x;
      entity = world[y][x].entity;
      if (entity->status == 2)
	{
	  get_color (entity->genoma);
	  GrPlotNC ((X_FIX + x), (Y_FIX + y), color);
	}
      else if ((entity->status == 1) && (entity->energy == 0))
	GrPlotNC ((X_FIX + x), (Y_FIX + y), GrBlack ());
    }
}


void
show_grx_stat ()
{
  old_pop = num_pop;
  num_pop = 0;
  num_active_genomi = 0;
  num_genomi = 0;
  max_dim = 0;
  min_dim = MAX_DIM;

  memset (num_dim, 0, sizeof (num_dim));
  memset (num_best_genoma, 0, sizeof (num_best_genoma));
  memset (best_genoma, 0, sizeof (best_genoma));

  for (genoma = first_genoma; genoma; genoma = genoma->next_genoma)
    {
      num_genomi++;
      if (genoma->num > 0)
	{
	  num_active_genomi++;
	  num_pop += genoma->num;
	  num_dim[(int) genoma->dim - 1] += genoma->num;
	  if (genoma->dim > max_dim)
	    max_dim = genoma->dim;
	  if (genoma->dim < min_dim)
	    min_dim = genoma->dim;
	  
	  j = NUM_BEST_GENOMA;
	  while ((genoma->num > num_best_genoma[j - 1]) && (j > 0))
	    {
	      num_best_genoma[j] = num_best_genoma[j - 1];
	      best_genoma[j] = best_genoma[j - 1];
	      j--;
	    }
	  if (j < NUM_BEST_GENOMA)
	    {
	      num_best_genoma[j] = genoma->num;
	      best_genoma[j] = genoma;
	    }
	}
    }

  sprintf (text, "WORLD  : %8lu  TIME   : %8lu",
	   world_num, world_time);
  GrTextXY (X_TEXT_FIX, Y_TEXT_FIX, text, GrWhite (), GrBlack ());
  sprintf (text, "POP    : %8lu  MAX DIM: %8u",
	   num_pop, max_dim);
  GrTextXY (X_TEXT_FIX, Y_TEXT_FIX + (11 * 1), text, GrWhite (), GrBlack ());
  sprintf (text, "OLD POP: %8lu  MIN DIM: %8u",
	   old_pop, min_dim);
  GrTextXY (X_TEXT_FIX, Y_TEXT_FIX + (11 * 2), text, GrWhite (), GrBlack ());
  sprintf (text, "DIF POP: %8ld  MAX ID : %8lX  NUM GEN: %6lu (%6lu)",
	   num_pop - old_pop, max_genoma_id, num_active_genomi, num_genomi);
  GrTextXY (X_TEXT_FIX, Y_TEXT_FIX + (11 * 3), text, GrWhite (), GrBlack ());

  strcpy (text, "## TYP         ID   PREVID DM      AGE MAXNUM    NUM");
  GrTextXY (X_TEXT_FIX, Y_TEXT_FIX + (11 * 5), text, GrWhite (), GrBlack ());

  for (j = 0; j < NUM_BEST_GENOMA; j++)
    {
      if (best_genoma[j])
	{
	  genoma = best_genoma[j];

	  num_mov = 0;
	  num_ene = 0;
	  num_atk = 0;
	  num_vir = 0;
	  num_imm = 0;
	  for (i = 0; i < genoma->dim; i++)
	    switch (genoma->gen[i])
	      {
	      case 1:
		num_mov++;
		break;
	      case 2:
		num_ene++;
		break;
	      case 50:
		num_vir++;
		break;
	      case 51:
		num_imm++;
		break;
	      case 52:
		num_atk++;
		break;
	      }
	  type_id[0] = 0;
	  if (num_ene > 0)
	    strcat(type_id, "P");
	  if (num_mov > 0)
	    strcat(type_id, "M");
	  if (num_atk > 0)
	    strcat(type_id, "A");
	  if (num_vir > 0)
	    strcat(type_id, "V");
	  if (num_imm > 0)
	    strcat(type_id, "I");
	  sprintf (text,
		   "%2d %-5s %8lX %8lX %2u %8lu %6lu %6lu",
		   j + 1,
		   type_id,
		   genoma->id,
		   genoma->prev_id,
		   genoma->dim,
		   world_time - genoma->birth_time,
		   genoma->max_num,
		   genoma->num);
	}
      else
	strcpy (text,
	  "                                                          ");
      get_color (genoma);
      GrTextXY (X_TEXT_FIX, Y_TEXT_FIX + (11 * (6 + j)), text,
		color, GrBlack ());
    }

  strcpy (text, "## GENOMA");
  GrTextXY (X_FIX, Y_TEXT_2_FIX, text, GrWhite (), GrBlack ());

  j = 0;
  i = 0;
  line = 0;
  do
    {
      if ((best_genoma[j])
	  && (i < best_genoma[j]->dim))
	{
	  if (i == 0)
	    sprintf (text, "%2d ", j + 1);
	  else
	    strcpy (text, "   ");
	  do
	    {
	      if (i < best_genoma[j]->dim)
		strcat (text, basm[best_genoma[j]->gen[i]]);
	      else
		strcat (text, "   ");
	      strcat (text, " ");
	      i++;
	    }
	  while ((i % 28) != 0);
	  get_color (best_genoma[j]);
	  GrTextXY (X_FIX, Y_TEXT_2_FIX + (11 * (line + 1)), text,
		    color, GrBlack ());
	  line++;
	}
      else
	{
	  j++;
	  i = 0;
	}
    }
  while ((line < 30) && (j < NUM_BEST_GENOMA));

  strcpy (text, "DM    NUM");
  GrTextXY (X_TEXT_2_FIX, Y_TEXT_FIX, text, GrWhite (), GrBlack ());

  for (j = 0; j < MAX_DIM; j++)
    {
      if (num_dim[j] > 0)
	{
	  sprintf (text, "%2d %6ld", j + 1, num_dim[j]);
	  color = MIN_COLOR +
	    (((j + 1) * STEP_COLOR) % (NUM_COLORS - MIN_COLOR));
	}
      else
	strcpy (text, "         ");
      GrTextXY (X_TEXT_2_FIX, Y_TEXT_FIX + (11 * (j + 1)),
		text,
		color, GrBlack ());
    }
}

void
evolve_world ()
{
  world_time++;

  for (coord = first_coord; coord; coord = coord->next_coord)
    {
      y = coord->y;
      x = coord->x;
      entity = world[y][x].entity;
      if (entity->energy > 0)
	if ((entity->status == 3) ||
	    (entity->time >= MAX_AGE + (entity->genoma->dim * MAX_GEN_AGE)))
	  {
	    entity->energy = 0;
	    entity->status = 1;
	    entity->genoma->num--;
	    if ((entity->genoma->num == 0)
		&& (entity->genoma->num_sup_genoma == 0))
	      del_genoma (entity->genoma);
	    if (entity->sub_genoma)
	      {
		entity->sub_genoma->num_sup_genoma--;
		if ((entity->sub_genoma->num == 0)
		    && (entity->sub_genoma->num_sup_genoma == 0))
		  del_genoma (entity->sub_genoma);
	      }
	  }
	else
	  entity->status = 0;
      else if (entity->status == 1)
	entity->status = 0;
      if ((entity->energy == 0)
	  && (entity->status == 0))
	{
	  free (entity);
	  world[y][x].entity = NULL;
	  del_curr_coord ();
	}
    }

  for (coord = first_coord; coord; coord = coord->next_coord)
    {
      y = coord->y;
      x = coord->x;
      entity = world[y][x].entity;
      if ((entity->energy > 0) && (entity->status == 0))
	{
	  genoma = entity->genoma;

	  entity->status = 1;

	  entity->time++;

	  eot = 0;
	  evolve_count = 0;


	  do
	    {
	      evolve_count++;

	      if (entity->gen_time >= genoma->dim)
		entity->gen_time -= genoma->dim;

	      switch (genoma->gen[(int) entity->gen_time++])
		{
		case 0:	/* NOP */
		  break;
		case 1:	/* MOV */
		  if ((entity->reg > 0)
		      && (entity->energy > MOV_ENERGY))
		    {
		      entity->energy -= MOV_ENERGY;
		      new_entity = get_near_empty (entity->reg, 1);
		      if (new_entity)
			{
			  entity->status = 2;
			  new_entity->status = 1;
			  world[y][x].entity = new_entity;
			  world[ny][nx].entity = entity;
			  y = ny;
			  x = nx;
			}
		    }
		  /*
		  eot = 1;
		  */
		  break;
		case 2:	/* ENE */
		  if (rnd (1000) < ENE_PROB)
		    {
		      if (entity->reg == 0)
			{
			  entity->energy += INC_ENERGY;
			  if (entity->energy > MAX_ENERGY)
			    entity->energy = MAX_ENERGY;
			}
		      else
			{
			  new_entity = get_near_full (entity->reg, 1);
			  if (new_entity)
			    {
			      new_entity->energy += INC_ENERGY;
			      if (new_entity->energy > MAX_ENERGY)
				new_entity->energy = MAX_ENERGY;
			    }
			}
		    }
		  eot = 1;
		  break;
		case 3:	/* REP */
		  if (entity->energy >
		      REP_ENERGY + entity->genoma->dim * REP_GEN_ENERGY)
		    {		/* REPRODUCTION */
		      if (entity->reg == 0)
			d0 = rnd (8) + 1;
		      else
			d0 = entity->reg;
		      d = d0;
		      dir = 2 * rnd (2) - 1;
		      do
			{
			  new_entity = get_near_empty (d, 1);
			  if (new_entity == NULL)
			    {
			      d += dir;
			      if (d > 8)
				d = 1;
			      else if (d < 1)
				d = 8;
			    }
			}
		      while ((d != d0) && (new_entity == NULL));
		      if (new_entity == NULL)
			entity->status = 3;
		      else
			{
			  entity->energy = START_ENERGY;
			  new_entity->energy = START_ENERGY;
			  new_entity->gen_time = 0;
			  new_entity->time = 0;
			  new_entity->mem = 0;
			  new_entity->reg = 0;
			  new_entity->status = 2;
			  new_entity->sub_genoma = NULL;
			  new_entity->sub_prob = 0;
			  if ((entity->sub_genoma)
			    && (rnd (1000) < entity->sub_prob))
			    genoma = entity->sub_genoma;
			  if (rnd (1000) >= MUT_PROB)
			    {
			      new_entity->genoma = genoma;
			      genoma->num++;
			      if (genoma->num > genoma->max_num)
				genoma->max_num = genoma->num;
			    }
			  else
			    {	/* MUTATION */
			      if ((new_genoma = malloc (sizeof (a_genoma))) == NULL)
				{
				  set_up_text_mode ();
				  printf ("End of memory...\n\n");
				  exit (1);
				}
			      new_genoma->dim = genoma->dim;
			      new_genoma->num_sup_genoma = 0;
			      memcpy (new_genoma->gen, genoma->gen, sizeof (a_gen) * new_genoma->dim);
			      switch (rnd (9))
				{
				case 0:	/* CHANGE */
				  pos = rnd (new_genoma->dim);
				  new_genoma->gen[pos] = rnd (BASM_DIM);
				  break;
				case 1:	/* INSERT */
				  if (new_genoma->dim < MAX_DIM)
				    {
				      pos = rnd (++new_genoma->dim);
				      for (j = new_genoma->dim; j >= pos + 1; j--)
					new_genoma->gen[j] = new_genoma->gen[j - 1];
				      new_genoma->gen[pos] = rnd (BASM_DIM);
				    }	/* if */
				  break;
				case 2:	/* DELETE */
				  if (new_genoma->dim > 1)
				    {
				      pos = rnd (new_genoma->dim--);
				      for (j = pos; j <= new_genoma->dim; j++)
					new_genoma->gen[j] = new_genoma->gen[j + 1];
				    }	/* if */
				  break;
				case 3:	/* MIX GENOMA */
				  d0 = rnd (8) + 1;
				  d = d0;
				  dir = 2 * rnd (2) - 1;
				  do
				    {
				      mut_entity = get_near_full (d, 1);
				      if ((!mut_entity)
					  || (mut_entity == new_entity)
					  || (mut_entity->genoma == entity->genoma))
					{
					  d += dir;
					  if (d > 8)
					    d = 1;
					  else if (d < 1)
					    d = 8;
					}
				    }
				  while ((!mut_entity) && (d != d0));
				  if ((mut_entity)
				      && (mut_entity != new_entity)
				    && (mut_entity->genoma != entity->genoma))
				    {
				      if (new_genoma->dim <= mut_entity->genoma->dim)
					i = new_genoma->dim;
				      else
					i = mut_entity->genoma->dim;
				      for (j = 0; j < i; j++)
					if (rnd (2) == 0)
					  new_genoma->gen[j] = mut_entity->genoma->gen[j];
				    }
				  break;
				case 4:	/* CONCAT GENOMA */
				  d0 = rnd (8) + 1;
				  d = d0;
				  dir = 2 * rnd (2) - 1;
				  do
				    {
				      mut_entity = get_near_full (d, 1);
				      if ((!mut_entity)
					  || (mut_entity == new_entity))
					{
					  d += dir;
					  if (d > 8)
					    d = 1;
					  else if (d < 1)
					    d = 8;
					}
				    }
				  while ((!mut_entity) && (d != d0));
				  if ((mut_entity)
				      && (mut_entity != new_entity))
				    if (rnd (2) == 1)
				      {
					i = new_genoma->dim + mut_entity->genoma->dim;
					if (i > MAX_DIM)
					  i = MAX_DIM;
					for (j = new_genoma->dim; j < i; j++)
					  new_genoma->gen[j] = mut_entity->genoma->gen[j - new_genoma->dim];
					new_genoma->dim = i;
				      }
				    else
				      {
					i = mut_entity->genoma->dim;
					if (i > (MAX_DIM - new_genoma->dim))
					  i = MAX_DIM - new_genoma->dim;
					for (j = 0; j < new_genoma->dim; j++)
					  new_genoma->gen[j + i] = new_genoma->gen[j];
					for (j = 0; j < i; j++)
					  new_genoma->gen[j] = mut_entity->genoma->gen[j];
				      }
				  break;
				case 5:	/* CUT GENOMA */
				  pos = rnd (new_genoma->dim);
				  i = rnd (new_genoma->dim - pos) + 1;
				  for (j = 0; j < i; j++)
				    new_genoma->gen[j] = new_genoma->gen[j + pos];
				  new_genoma->dim = i;
				  break;
				case 6:	/* REDUCE GENOMA */
				  i = 0;
				  for (j = 0; j < new_genoma->dim; j++)
				    {
				      if (rnd (2) == 0)
					{
					  new_genoma->gen[i] = new_genoma->gen[j];
					  i++;
					}
				    }
				  if (i != 0)
				    new_genoma->dim = i;
				  else
				    new_genoma->dim = 1;
				  break;
				}	/* switch */
			      if ((genoma->dim == new_genoma->dim)
				  &&(!memcmp(genoma->gen, new_genoma->gen, genoma->dim * sizeof(a_gen))))
				{
				  free(new_genoma);
				  new_entity->genoma = genoma;
				  genoma->num++;
				  if (genoma->num > genoma->max_num)
				    genoma->max_num = genoma->num;
				}
			      else
				{
				  new_entity->genoma = new_genoma;
				  last_genoma->next_genoma = new_genoma;
				  new_genoma->prev_genoma = last_genoma;
				  new_genoma->next_genoma = NULL;
				  last_genoma = new_genoma;
				  new_genoma->id = ++max_genoma_id;
				  new_genoma->prev_id = genoma->id;
				  new_genoma->birth_time = world_time;
				  new_genoma->num = 1;
				  new_genoma->max_num = 1;
				}
			    }	/* if mutation */
			}
		    }
		  genoma = entity->genoma;
		  eot = 1;
		  break;
		case 4:	/* INF */
		  for (dist = 1; dist <= MAX_DIST; dist++)
		    for (d = 1; d <= 8; d++)
		      {
			new_entity = get_near_full (d, dist);
			if (new_entity)
			  new_entity->reg = entity->reg;
		      }
		  break;
		case 5:	/* LFN */
		  if (entity->energy > LXX_ENERGY)
		    {
		      if (entity->reg == 0)
			d0 = rnd (8) + 1;
		      else
			d0 = entity->reg;
		      entity->reg = 0;
		      d = d0;
		      dist = 1;
		      dir = 2 * rnd (2) - 1;
		      do
			{
			  new_entity = get_near_empty (d, dist);
			  if (new_entity)
			    entity->reg = d;
			  else
			    {
			      d += dir;
			      if (d > 8)
				d = 1;
			      else if (d < 1)
				d = 8;
			      if (d == d0)
				dist++;
			    }
			}
		      while ((entity->reg == 0) && (dist <= MAX_DIST));
		    }
		  break;
		case 6:	/* LFF */
		  if (entity->energy > LXX_ENERGY)
		    {
		      entity->energy -= LXX_ENERGY;
		      if (entity->reg == 0)
			d0 = rnd (8) + 1;
		      else
			d0 = entity->reg;
		      entity->reg = 0;
		      d = d0;
		      dist = 1;
		      dir = 2 * rnd (2) - 1;
		      do
			{
			  new_entity = get_near_full (d, dist);
			  if (new_entity)
			    entity->reg = d;
			  else
			    {
			      d += dir;
			      if (d > 8)
				d = 1;
			      else if (d < 1)
				d = 8;
			      if (d == d0)
				dist++;
			    }
			}
		      while ((entity->reg == 0) && (dist <= MAX_DIST));
		    }
		  break;
		case 7:	/* LFD */
		  if (entity->energy > LXX_ENERGY)
		    {
		      entity->energy -= LXX_ENERGY;
		      if (entity->reg == 0)
			d0 = rnd (8) + 1;
		      else
			d0 = entity->reg;
		      entity->reg = 0;
		      d = d0;
		      dist = 1;
		      dir = 2 * rnd (2) - 1;
		      do
			{
			  new_entity = get_near_full (d, dist);
			  if ((new_entity)
			      && (new_entity->genoma->dim != genoma->dim))
			    entity->reg = d;
			  else
			    {
			      d += dir;
			      if (d > 8)
				d = 1;
			      else if (d < 1)
				d = 8;
			      if (d == d0)
				dist++;
			    }
			}
		      while ((entity->reg == 0) && (dist <= MAX_DIST));
		    }
		  break;
		case 8:	/* LFS */
		  if (entity->energy > LXX_ENERGY)
		    {
		      entity->energy -= LXX_ENERGY;
		      if (entity->reg == 0)
			d0 = rnd (8) + 1;
		      else
			d0 = entity->reg;
		      entity->reg = 0;
		      d = d0;
		      dist = 1;
		      dir = 2 * rnd (2) - 1;
		      do
			{
			  new_entity = get_near_full (d, dist);
			  if ((new_entity)
			      && (new_entity->genoma->dim < genoma->dim))
			    entity->reg = d;
			  else
			    {
			      d += dir;
			      if (d > 8)
				d = 1;
			      else if (d < 1)
				d = 8;
			      if (d == d0)
				dist++;
			    }
			}
		      while ((entity->reg == 0) && (dist <= MAX_DIST));
		    }
		  break;
		case 9:	/* LFB */
		  if (entity->energy > LXX_ENERGY)
		    {
		      entity->energy -= LXX_ENERGY;
		      if (entity->reg == 0)
			d0 = rnd (8) + 1;
		      else
			d0 = entity->reg;
		      entity->reg = 0;
		      d = d0;
		      dist = 1;
		      dir = 2 * rnd (2) - 1;
		      do
			{
			  new_entity = get_near_full (d, dist);
			  if ((new_entity)
			      && (new_entity->genoma->dim > genoma->dim))
			    entity->reg = d;
			  else
			    {
			      d += dir;
			      if (d > 8)
				d = 1;
			      else if (d < 1)
				d = 8;
			      if (d == d0)
				dist++;
			    }
			}
		      while ((entity->reg == 0) && (dist <= MAX_DIST));
		    }
		  break;
		case 10:	/* LFE */
		  if (entity->energy > LXX_ENERGY)
		    {
		      entity->energy -= LXX_ENERGY;
		      if (entity->reg == 0)
			d0 = rnd (8) + 1;
		      else
			d0 = entity->reg;
		      entity->reg = 0;
		      d = d0;
		      dist = 1;
		      dir = 2 * rnd (2) - 1;
		      do
			{
			  new_entity = get_near_full (d, dist);
			  if ((new_entity)
			      && (new_entity->genoma->dim == genoma->dim))
			    entity->reg = d;
			  else
			    {
			      d += dir;
			      if (d > 8)
				d = 1;
			      else if (d < 1)
				d = 8;
			      if (d == d0)
				dist++;
			    }
			}
		      while ((entity->reg == 0) && (dist <= MAX_DIST));
		    }
		  break;
		case 11:	/* LEI */
		  if (entity->energy > LXX_ENERGY)
		    {
		      entity->energy -= LXX_ENERGY;
		      if (entity->reg == 0)
			d0 = rnd (8) + 1;
		      else
			d0 = entity->reg;
		      entity->reg = 0;
		      d = d0;
		      dist = 1;
		      dir = 2 * rnd (2) - 1;
		      do
			{
			  new_entity = get_near_full (d, dist);
			  /*
			  if ((new_entity)
			      && ((new_entity->genoma->id == genoma->id)
			     || (new_entity->genoma->prev_id == genoma->id)))
			     */
			  if ((new_entity)
			      && (new_entity->genoma->id == genoma->id))
			    entity->reg = d;
			  else
			    {
			      d += dir;
			      if (d > 8)
				d = 1;
			      else if (d < 1)
				d = 8;
			      if (d == d0)
				dist++;
			    }
			}
		      while ((entity->reg == 0) && (dist <= MAX_DIST));
		    }
		  break;
		case 12:	/* LDI */
		  if (entity->energy > LXX_ENERGY)
		    {
		      entity->energy -= LXX_ENERGY;
		      if (entity->reg == 0)
			d0 = rnd (8) + 1;
		      else
			d0 = entity->reg;
		      entity->reg = 0;
		      d = d0;
		      dist = 1;
		      dir = 2 * rnd (2) - 1;
		      do
			{
			  new_entity = get_near_full (d, dist);
			  /*
			    if ((new_entity)
			      && (new_entity->genoma->id != genoma->id)
			      && (new_entity->genoma->prev_id != genoma->id))
			      */
			  if ((new_entity)
			      && (new_entity->genoma->id != genoma->id))
			    entity->reg = d;
			  else
			    {
			      d += dir;
			      if (d > 8)
				d = 1;
			      else if (d < 1)
				d = 8;
			      if (d == d0)
				dist++;
			    }
			}
		      while ((entity->reg == 0) && (dist <= MAX_DIST));
		    }
		  break;
		case 13:	/* LFY */
		  if (entity->energy > LXX_ENERGY)
		    {
		      entity->energy -= LXX_ENERGY;
		      if (entity->reg == 0)
			d0 = rnd (8) + 1;
		      else
			d0 = entity->reg;
		      entity->reg = 0;
		      d = d0;
		      dist = 1;
		      dir = 2 * rnd (2) - 1;
		      do
			{
			  new_entity = get_near_full (d, dist);
			  if ((new_entity)
			      && (new_entity->time < entity->time))
			    entity->reg = d;
			  else
			    {
			      d += dir;
			      if (d > 8)
				d = 1;
			      else if (d < 1)
				d = 8;
			      if (d == d0)
				dist++;
			    }
			}
		      while ((entity->reg == 0) && (dist <= MAX_DIST));
		    }
		  break;
		case 14:	/* LFO */
		  if (entity->energy > LXX_ENERGY)
		    {
		      entity->energy -= LXX_ENERGY;
		      if (entity->reg == 0)
			d0 = rnd (8) + 1;
		      else
			d0 = entity->reg;
		      entity->reg = 0;
		      d = d0;
		      dist = 1;
		      dir = 2 * rnd (2) - 1;
		      do
			{
			  new_entity = get_near_full (d, dist);
			  if ((new_entity)
			      && (new_entity->time > entity->time))
			    entity->reg = d;
			  else
			    {
			      d += dir;
			      if (d > 8)
				d = 1;
			      else if (d < 1)
				d = 8;
			      if (d == d0)
				dist++;

			    }
			}
		      while ((entity->reg == 0) && (dist <= MAX_DIST));
		    }
		  break;
		case 15:	/* CFN */
		  if (entity->energy > CXX_ENERGY)
		    {
		      entity->energy -= CXX_ENERGY;
		      if (entity->reg != 0)
			{
			  dist = 1;
			  d = entity->reg;
			  entity->reg = 0;
			  while ((entity->reg == 0) && (dist <= MAX_DIST))
			    {
			      new_entity = get_near_empty (d, dist);
			      if (new_entity)
				entity->reg = d;
			      else
				dist++;
			    }
			}
		    }
		  break;
		case 16:	/* CFF */
		  if (entity->energy > CXX_ENERGY)
		    {
		      entity->energy -= CXX_ENERGY;
		      if (entity->reg != 0)
			{
			  dist = 1;
			  d = entity->reg;
			  entity->reg = 0;
			  while ((entity->reg == 0) && (dist <= MAX_DIST))
			    {
			      new_entity = get_near_full (d, dist);
			      if (new_entity)
				entity->reg = d;
			      else
				dist++;
			    }
			}
		    }
		  break;
		case 17:	/* CFD */
		  if (entity->energy > CXX_ENERGY)
		    {
		      entity->energy -= CXX_ENERGY;
		      if (entity->reg != 0)
			{
			  dist = 1;
			  d = entity->reg;
			  entity->reg = 0;
			  while ((entity->reg == 0) && (dist <= MAX_DIST))
			    {
			      new_entity = get_near_full (d, dist);
			      if ((new_entity)
				  && (new_entity->genoma->dim != genoma->dim))
				entity->reg = d;
			      else
				dist++;
			    }
			}
		    }
		  break;
		case 18:	/* CFS */
		  if (entity->energy > CXX_ENERGY)
		    {
		      entity->energy -= CXX_ENERGY;
		      if (entity->reg != 0)
			{
			  dist = 1;
			  d = entity->reg;
			  entity->reg = 0;
			  while ((entity->reg == 0) && (dist <= MAX_DIST))
			    {
			      new_entity = get_near_full (d, dist);
			      if ((new_entity)
				  && (new_entity->genoma->dim < genoma->dim))
				entity->reg = d;
			      else
				dist++;
			    }
			}
		    }
		  break;
		case 19:	/* CFB */
		  if (entity->energy > CXX_ENERGY)
		    {
		      entity->energy -= CXX_ENERGY;
		      if (entity->reg != 0)
			{
			  dist = 1;
			  d = entity->reg;
			  entity->reg = 0;
			  while ((entity->reg == 0) && (dist <= MAX_DIST))
			    {
			      new_entity = get_near_full (d, dist);
			      if ((new_entity)
				  && (new_entity->genoma->dim > genoma->dim))
				entity->reg = d;
			      else
				dist++;
			    }
			}
		    }
		  break;
		case 20:	/* CFE */
		  if (entity->energy > CXX_ENERGY)
		    {
		      entity->energy -= CXX_ENERGY;
		      if (entity->reg != 0)
			{
			  dist = 1;
			  d = entity->reg;
			  entity->reg = 0;
			  while ((entity->reg == 0) && (dist <= MAX_DIST))
			    {
			      new_entity = get_near_full (d, dist);
			      if ((new_entity)
				  && (new_entity->genoma->dim == genoma->dim))
				entity->reg = d;
			      else
				dist++;
			    }
			}
		    }
		  break;
		case 21:	/* CEI */
		  if (entity->energy > CXX_ENERGY)
		    {
		      entity->energy -= CXX_ENERGY;
		      if (entity->reg != 0)
			{
			  dist = 1;
			  d = entity->reg;
			  entity->reg = 0;
			  while ((entity->reg == 0) && (dist <= MAX_DIST))
			    {
			      new_entity = get_near_full (d, dist);
			      /*
				if ((new_entity)
				  && ((new_entity->genoma->id == genoma->id)
			      || (new_entity->genoma->prev_id == genoma->id)))
			      */
			      if ((new_entity)
				  && (new_entity->genoma->id == genoma->id))
				entity->reg = d;
			      else
				dist++;
			    }
			}
		    }
		  break;
		case 22:	/* CDI */
		  if (entity->energy > CXX_ENERGY)
		    {
		      entity->energy -= CXX_ENERGY;
		      if (entity->reg != 0)
			{
			  dist = 1;
			  d = entity->reg;
			  entity->reg = 0;
			  while ((entity->reg == 0) && (dist <= MAX_DIST))
			    {
			      new_entity = get_near_full (d, dist);
			      /*
				if ((new_entity)
				  && (new_entity->genoma->id != genoma->id)
			      && (new_entity->genoma->prev_id != genoma->id))
			      */
				if ((new_entity)
				  && (new_entity->genoma->id != genoma->id))
				entity->reg = d;
			      else
				dist++;
			    }
			}
		    }
		  break;
		case 23:	/* CFY */
		  if (entity->energy > CXX_ENERGY)
		    {
		      entity->energy -= CXX_ENERGY;
		      if (entity->reg != 0)
			{
			  dist = 1;
			  d = entity->reg;
			  entity->reg = 0;
			  while ((entity->reg == 0) && (dist <= MAX_DIST))
			    {
			      new_entity = get_near_full (d, dist);
			      if ((new_entity)
				  && (new_entity->time < entity->time))
				entity->reg = d;
			      else
				dist++;
			    }
			}
		    }
		  break;
		case 24:	/* CFO */
		  if (entity->energy > CXX_ENERGY)
		    {
		      entity->energy -= CXX_ENERGY;
		      if (entity->reg != 0)
			{
			  dist = 1;
			  d = entity->reg;
			  entity->reg = 0;
			  while ((entity->reg == 0) && (dist <= MAX_DIST))
			    {
			      new_entity = get_near_full (d, dist);
			      if ((new_entity)
				  && (new_entity->time > entity->time))
				entity->reg = d;
			      else
				dist++;
			    }
			}
		    }
		  break;
		case 25:	/* RND */
		  entity->reg = rnd (9);
		  break;
		case 26:	/* IF0 */
		  if (entity->reg != 0)
		    {
		      do
			entity->gen_time++;
		      while((entity->gen_time < genoma->dim)
			    && (genoma->gen[entity->gen_time - 1] != 53)
			    && (genoma->gen[entity->gen_time - 1] != 54));
		    }
		  break;
		case 27:	/* IFN */
		  if (entity->reg == 0)
		    {
		      do
			entity->gen_time++;
		      while((entity->gen_time < genoma->dim)
			    && (genoma->gen[entity->gen_time - 1] != 53)
			    && (genoma->gen[entity->gen_time - 1] != 54));
		    }
		  break;
		case 28:	/* IFE */
		  if (entity->reg != entity->mem)
		    {
		      do
			entity->gen_time++;
		      while((entity->gen_time < genoma->dim)
			    && (genoma->gen[entity->gen_time - 1] != 53)
			    && (genoma->gen[entity->gen_time - 1] != 54));
		    }
		  break;
		case 29:	/* IFD */
		  if (entity->reg == entity->mem)
		    {
		      do
			entity->gen_time++;
		      while((entity->gen_time < genoma->dim)
			    && (genoma->gen[entity->gen_time - 1] != 53)
			    && (genoma->gen[entity->gen_time - 1] != 54));
		    }
		  break;
		case 30:	/* TRN */
		  if (entity->reg != 0)
		    entity->reg = 9 - entity->reg;
		  break;
		case 31:	/* TRR */
		  if (entity->reg != 0)
		    {
		      entity->reg += 1;
		      if (entity->reg > 8)
			entity->reg -= 8;
		    }
		  break;
		case 32:	/* TRL */
		  if (entity->reg != 0)
		    {
		      entity->reg -= 1;
		      if (entity->reg < 1)
			entity->reg += 8;
		    }
		  break;
		case 33:	/* GOF */
		  entity->gen_time += entity->reg;
		  while (entity->gen_time >= genoma->dim)
		    entity->gen_time -= genoma->dim;
		  break;
		case 34:	/* GOB */
		  entity->gen_time -= entity->reg;
		  while (entity->gen_time < 0)
		    entity->gen_time += genoma->dim;
		  break;
		case 35:	/* TO0 */
		  entity->reg = 0;
		  break;
		case 36:	/* TO1 */
		  entity->reg = 1;
		  break;
		case 37:	/* TO2 */
		  entity->reg = 2;
		  break;
		case 38:	/* TO3 */
		  entity->reg = 3;
		  break;
		case 39:	/* TO4 */
		  entity->reg = 4;
		  break;
		case 40:	/* TO5 */
		  entity->reg = 5;
		  break;
		case 41:	/* TO6 */
		  entity->reg = 6;
		  break;
		case 42:	/* TO7 */
		  entity->reg = 7;
		  break;
		case 43:	/* TO8 */
		  entity->reg = 8;
		  break;
		case 44:	/* EOT */
		  eot = 1;
		  break;
		case 45:	/* SWP */
		  wk_mem = entity->mem;
		  entity->mem = entity->reg;
		  entity->reg = wk_mem;
		  break;
		case 46:	/* WME */
		  entity->mem = entity->reg;
		  break;
		case 47:	/* RMW */
		  entity->reg = entity->mem;
		  break;
		case 48:	/* WSI */
		  world[y][x].signal = entity->reg;
		  break;
		case 49:	/* RSI */
		  entity->reg = world[y][x].signal;
		  break;
		case 50:	/* VIR */
		  if ((entity->reg > 0)
		      && (entity->energy > VIR_ENERGY))
		    {
		      entity->energy -= VIR_ENERGY;
		      mut_entity = get_near_full (entity->reg, 1);
		      if ((mut_entity)
			  && (mut_entity->genoma != genoma)
			  && (mut_entity->sub_genoma != genoma))
			{
			  if (mut_entity->sub_genoma)
			    {
			      mut_entity->sub_genoma->num_sup_genoma--;
			      if ((mut_entity->sub_genoma->num == 0)
				  && (mut_entity->sub_genoma->num_sup_genoma == 0))
				del_genoma (mut_entity->sub_genoma);
			    }
			  mut_entity->sub_genoma = genoma;
			  mut_entity->sub_prob = rnd(1001);
			  genoma->num_sup_genoma++;
			}
		    }
		  eot = 1;
		  break;
		case 51:	/* IMM */
		  if (entity->energy > IMM_ENERGY)
		    {
		      entity->energy -= IMM_ENERGY;
		      if (entity->sub_genoma)
			{
			  entity->sub_genoma->num_sup_genoma--;
			  if ((entity->sub_genoma->num == 0)
			      && (entity->sub_genoma->num_sup_genoma == 0))
			    del_genoma (entity->sub_genoma);
			  entity->sub_genoma = NULL;
			  entity->sub_prob = 0;
			}
		    }
		  eot = 1;
		  break;
		case 52:	/* ATK */
		  if ((entity->reg > 0)
		      && (entity->energy > ATK_ENERGY))
		    {
		      entity->energy -= ATK_ENERGY;
		      new_entity = get_near (entity->reg, 1);
		      if (new_entity->energy > ATK_GET_ENERGY)
			{
			  new_entity->energy -= ATK_GET_ENERGY;
			  new_entity->reg = 9 - entity->reg;
			  entity->energy += ATK_GET_ENERGY;
			  if (entity->energy > MAX_ENERGY)
			    entity->energy = MAX_ENERGY;
			}
		      else
			{
			  if (new_entity->energy > 0)
			    {
			      entity->energy += new_entity->energy;
			      if (entity->energy > MAX_ENERGY)
				entity->energy = MAX_ENERGY;
			      new_entity->genoma->num--;
			      if ((new_entity->genoma->num == 0)
				  && (new_entity->genoma->num_sup_genoma == 0))
				del_genoma (new_entity->genoma);
			      if (new_entity->sub_genoma)
				{
				  new_entity->sub_genoma->num_sup_genoma--;
				  if ((new_entity->sub_genoma->num == 0)
				      && (new_entity->sub_genoma->num_sup_genoma == 0))
				    del_genoma (new_entity->sub_genoma);
				}
			      new_entity->energy = 0;
			    }
			  entity->status = 2;
			  new_entity->status = 1;
			  world[y][x].entity = new_entity;
			  world[ny][nx].entity = entity;
			  y = ny;
			  x = nx;
			}
		    }
		  eot = 1;
		  break;
		case 53:	/* ELS */
		  do
		    entity->gen_time++;
		  while((entity->gen_time < genoma->dim)
			&& (genoma->gen[entity->gen_time - 1] != 54));
		  break;
		case 54:	/* EIF */
		  /* DO NOTHING... */
		  break;
		}
	    }
	  while ((eot == 0) && (evolve_count < genoma->dim));
	}
    }
}


void
set_options ()
{
  GrClearScreen (GrBlack ());

  GrBoxNC (0, 0, X_SCR - 1, Y_SCR - 1, GrWhite ());
  strcpy (text, "SIMEV 0.1 - OPTIONS");
  GrTextXY (X_FIX, 2, text, GrWhite (), GrBlack ());

  for (j = 0; j <NUM_OPTIONS; j++)
    show_option(j);

  ask_option(3);

  clear_graph_mode ();
  show_grx_world_all ();
  show_grx_stat ();
}

/*  
int inc_energy = INC_ENERGY,
  start_energy = START_ENERGY,
  rep_energy = REP_ENERGY,
  rep_gen_energy = REP_GEN_ENERGY,
  mov_energy = MOV_ENERGY,
  atk_energy = ATK_ENERGY,
  atk_get_energy = ATK_GET_ENERGY,
  lxx_energy = LXX_ENERGY,
  cxx_energy = CXX_ENERGY,
  vir_energy = VIR_ENERGY,
  imm_energy = IMM_ENERGY,
  mut_prob = MUT_PROB,
  ene_prob = ENE_PROB,
  max_age = MAX_AGE,
  max_gen_age = MAX_GEN_AGE,
  max_dist = MAX_DIST;
*/


void
snapshot_world ()
{
  FILE *file;

  if ((file = fopen ("simev.snp", "w")) == NULL)
    {
      set_up_text_mode ();
      printf ("Impossibile aprire in scrittura il file SIMEV.SNP\n");
      exit (1);
    }

  /*  for(x=0; x<X_DIM; x++)
     fprintf(file, "-");
     fprintf(file, "\n"); */

  for (y = 0; y < Y_DIM; y++)
    {
      for (x = 0; x < X_DIM; x++)
	{
	  entity = world[y][x].entity;
	  if (entity)
	    if (entity->energy > 0)
	      fprintf (file, "%c", ch[(int) entity->genoma->dim]);
	    else if (entity->energy > 0)
	      fprintf (file, " ");
	    else
	      fprintf (file, "*");
	}
      fprintf (file, "\n");
    }

  fclose (file);

  if ((file = fopen ("simev.out", "w")) == NULL)
    {
      set_up_text_mode ();
      printf ("Impossibile aprire in scrittura il file SIMEV.OUT\n");
      exit (1);
    }

  for (coord = first_coord; coord; coord = coord->next_coord)
    {
      y = coord->y;
      x = coord->x;
      entity = world[y][x].entity;
      if (entity)
	if (entity->energy > 0)
	  {
	    fprintf (file, "(%3i,%3i) [%10lu] (%10lu) %8lu %3i %3i %1i", x, y,
		     entity->genoma->id,
		     entity->genoma->prev_id,
		     entity->genoma->num,
		     entity->genoma->dim,
		     entity->energy, entity->mem);
	    for (j = 0; j < entity->genoma->dim; j++)
	      fprintf (file, " %3s", basm[entity->genoma->gen[j]]);
	    fprintf (file, "\n");
	  }
    }

  fclose (file);

}

void
main ()
{
  int stime;
  long ltime;

  initscr ();
  cbreak ();
  noecho ();
  nodelay (stdscr, TRUE);
  keypad (stdscr, TRUE);

  if ((genfile = fopen ("simevgen.dat", "w")) == NULL)
    {
      set_up_text_mode ();
      printf ("Impossibile aprire in scrittura il file SIMEVGEN.DAT\n");
      exit (1);
    }

  /* Use time to get a seed for random numbers */

  ltime = time (NULL);
  stime = (unsigned int) ltime / 2;
  srand (stime);

  set_up_graph_mode ();

  init_world ();
  fill_world ();

  count_stat_rate = update_stat_rate;

  for (;;)
    {

      if ((single_step_mode)&&(c != ' '))
	do
	  c = getch();
	while (c == ERR);
      else
      do
	{
	  if (show_world)
	    show_grx_world ();

	  if (show_stat)
	    if (count_stat_rate < update_stat_rate)
	      count_stat_rate++;
	    else
	      {
		count_stat_rate = 1;
		show_grx_stat ();
	      }

	  if (first_genoma == NULL)
	    {
	      world_num++;
	      world_time = 0;
	      max_genoma_id = 0;
	      old_pop = 0;
	      count_stat_rate = 1;
	      init_world ();
	      fill_world ();
	      if (show_world)
		show_grx_world_all ();
	    }

	  evolve_world ();

	  c = getch();
	}
      while ((!single_step_mode)&&(c == ERR));

      switch (toupper (c))
	{
	case 'C':
	  color_type = 1 - color_type;
	  if (show_world)
	    show_grx_world_all ();
	  if (show_stat)
	    show_grx_stat ();
	  break;
	case '1':
	  update_stat_rate = 1;
	  break;
	case '2':
	  update_stat_rate = 5;
	  break;
	case '3':
	  update_stat_rate = 10;
	  break;
	case '4':
	  update_stat_rate = 50;
	  break;
	case '5':
	  update_stat_rate = 100;
	  break;
	case '6':
	  update_stat_rate = 500;
	  break;
	case '7':
	  update_stat_rate = 1000;
	  break;
	case 'R':
	  set_up_graph_mode ();
	  show_grx_world_all ();
	  show_grx_stat ();
	  break;
	case 'S':
	  show_stat = !show_stat;
	  break;
	case 'D':
	  show_world = 0;
	  show_stat = 0;
	  set_up_text_mode ();
	  break;
	case 'W':
	  switch (show_world)
	    {
	    case 1:
	      show_world = 0;
	      break;
	    case 0:
	      show_grx_world_all ();
	      show_world = 1;
	      break;
	    }
	  break;
	case 'F':
	  snapshot_world ();
	  break;
	case 'Q':
	  exit_simev (0);
	  break;
	case 'P':
	  single_step_mode = 1 - single_step_mode;
	  break;
	  case 'O':
	    set_options ();
	    break;
	}
    }
}

